/*
 * gyro.c
 *
 * IMPORTANT: This project uses i2cmaster.S.
 * Make sure that twimaster.c is NOT in this project's folder.
 *
 * Add float libraries and the following switches to Linker Options:
 *    -Wl,-u,vfscanf -lscanf_flt -lm
 *    -Wl,-u,vfscanf -lscanf_flt -lm
 *
 * Created: 11/23/2014 11:29:45 AM
 *  Author: Shela Wang and Christine Soong
 * 
 * Credit:
 * This I2C code was based on Brian Harding's code from ECE 4760 Spring 2011.
 */ 

#define F_CPU 16000000UL
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <math.h>

#include "uart.h"
#include "i2cmaster.h"

//Debug flag
//Set to 0 if using UART to send over Bluetooth; 1 if using UART for PuTTY
#define DEBUG 0

//Bit macros
#define set_bit(address,bit) (address |= (1<<bit))
#define clear_bit(address,bit) (address &= ~(1<<bit))
#define toggle_bit(address,bit) (address ^= (1<<bit))
#define check_bit(address,bit) ((address & (1<<bit)) == (1<<bit))

//MPU-6050 registers
#define MPU_6050   (0x68<<1) //device address of MPU-6050; see datasheet
#define WHO_AM_I    0x75 //Register 117, holds upper 6 bits of MPU6050's 7-bit I2C address = 0x68
#define PWR_MGMT_1  0x6B //Register 107, used to wake the MPU from sleep mode
#define INT_ENABLE	0x38 //Register 56, used to enable interrupts on the MPU
#define INT_STATUS	0x3A //Register 58, the location of the MPU's interrupt flags

#define GYRO_CONFIG	0x1B //Register 26, used to trigger gyroscope self-test & configure gyroscopes' full scale range
#define FS_SEL_0	0x00 //FS_SEL[1:0] - 0 = 250 o/s 
#define FS_SEL_1	0x08 //FS_SEL[1:0] - 1 = 500 o/s
#define FS_SEL_2	0x10 //FS_SEL[1:0] - 2 = 1000 o/s 
#define FS_SEL_3	0x18 //FS_SEL[1:0] - 3 = 2000 o/s

#define GYRO_XOUT_H 0x43 //Register 67, contains [15:8] of 16 bit 2's complement value of most recent x axis gyroscope measurement
#define GYRO_XOUT_L 0x44 //Register 68, contains [7:0]  of 16 bit 2's complement value of most recent x axis gyroscope measurement
#define GYRO_YOUT_H 0x45 //Register 69, contains [15:8] of 16 bit 2's complement value of most recent y axis gyroscope measurement
#define GYRO_YOUT_L 0x46 //Register 70, contains [7:0]  of 16 bit 2's complement value of most recent y axis gyroscope measurement
#define GYRO_ZOUT_H 0x47 //Register 71, contains [15:8] of 16 bit 2's complement value of most recent z axis gyroscope measurement
#define GYRO_ZOUT_L 0x48 //Register 72, contains [7:0]  of 16 bit 2's complement value of most recent z axis gyroscope measurement

#define ACCEL_XOUT_H 0x3B
#define ACCEL_XOUT_L 0x3C
#define ACCEL_YOUT_H 0x3D
#define ACCEL_YOUT_L 0x3E
#define ACCEL_ZOUT_H 0x3F
#define ACCEL_ZOUT_L 0x40

//UART setup
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//Gyroscope measurements (in degrees/sec)
double gyro_x = 0;
double gyro_y = 0;
double gyro_z = 0;

//Raw measured gyro values
int16_t gyro_x_raw;
int16_t gyro_y_raw;
int16_t gyro_z_raw;

//Used for measuring gyro data
unsigned char data_ready = 0;
unsigned char gyro_x_upper;
unsigned char gyro_x_lower;
unsigned char gyro_y_upper;
unsigned char gyro_y_lower;
unsigned char gyro_z_upper;
unsigned char gyro_z_lower;

//Initial gyroscope values (gyro has offsets on startup)
int16_t gyro_x_offset;
int16_t gyro_y_offset;
int16_t gyro_z_offset;

//Absolute angles of the gyroscope
double x_angle = 0;
double y_angle = 0;

//Raw measured accel values
int16_t accel_x_raw;
int16_t accel_y_raw;
int16_t accel_z_raw;

unsigned char accel_x_upper;
unsigned char accel_x_lower;
unsigned char accel_y_upper;
unsigned char accel_y_lower;
unsigned char accel_z_upper;
unsigned char accel_z_lower;

char buttonPressed = 0;


//************ MPU HELPER FUNCTIONS ***************

/*
 * Reads a byte from register regNum in the MPU.
 * Params:
 *		regNum - the register on the MPU to read from
 * Returns: the data read from register regNum
 */
unsigned char mpu_read(unsigned char regNum) {
	unsigned char data;
	unsigned char f; //failbit

	//set device address and write mode
	f = i2c_start(MPU_6050+I2C_WRITE);
	if (f && DEBUG) fprintf(stderr, "mpu_read i2c_start (W) failbit: %d\n", f);

	//set the read address
	f = i2c_write(regNum);
	if (f && DEBUG) fprintf(stderr, "mpu_read i2c_write failbit: %d\n", f);
	i2c_stop();

	//set device address and read mode
	f = i2c_start(MPU_6050+I2C_READ);
	if (f && DEBUG) fprintf(stderr, "mpu_read i2c_start (R) failbit: %d\n", f);
	
	//read one byte of data
	data = i2c_readNak();

	//set stop conditon = release bus
	i2c_stop();

	return data;
}

/*
 * Writes a byte (value) to register regNum in the MPU.
 * Params:
 *		regNum - the register number to write to on the MPU
 *		value  - the value to write to the specified register
 * Returns: 0 if write was successful; 1 if it was unsuccessful
 */
unsigned char mpu_write(unsigned char regNum, unsigned char value){
	unsigned char f; //failbit

	f = i2c_start(MPU_6050+I2C_WRITE);	//set device address and write mode
	if (f) {
		if (DEBUG) fprintf(stderr, "mpu_write i2c_start (W) failbit: %d\n", f);
		return f;
	}

	f = i2c_write(regNum);	//register address
	if (f) {
		if (DEBUG) fprintf(stderr, "mpu_write i2c_start (R) failbit: %d\n", f);
		return f;
	}

	f = i2c_write(value);	//writing value to register
	if (f) {
		if (DEBUG) fprintf(stderr, "mpu_write i2c_start (WriteData) failbit: %d\n", f);
		return f;
	}

	//set stop conditon = release bus
	i2c_stop();
	return 0;
}

/*
 * Concatenates the upper 8 bits & lower 8 bits into a 16-bit signed value.
 * Params:
 *		upper - the upper 8 bits
 *		lower - the lower 8 bits
 * Returns: a 16-bit signed int
 */
int16_t combine(unsigned char upper, unsigned char lower){
	int16_t result = 0;
	result = ((upper << 8) | lower);
	return result;
}

/*
 * Reads in the raw x, y, and z values of the gyroscope and puts them
 * in global variables gyro_*_raw.
 *
 * NOTE: This function does NOT check the Data Ready bit before reading
 * gyro data. It is up to the caller to do that before calling this.
 *
 * Returns: nothing
 */
void readRawGyroData() {
	gyro_x_upper = mpu_read(GYRO_XOUT_H);
    gyro_x_lower = mpu_read(GYRO_XOUT_L);
    gyro_x_raw   = combine(gyro_x_upper, gyro_x_lower);
    
    gyro_y_upper = mpu_read(GYRO_YOUT_H);
    gyro_y_lower = mpu_read(GYRO_YOUT_L);
    gyro_y_raw   = combine(gyro_y_upper, gyro_y_lower);
    
    gyro_z_upper = mpu_read(GYRO_ZOUT_H);
    gyro_z_lower = mpu_read(GYRO_ZOUT_L);
    gyro_z_raw   = combine(gyro_z_upper, gyro_z_lower);
}

/*
 * Takes in raw gyroscope x, y, and z values, converts them to actual measurements
 * in degrees/sec, and puts the values in global variables gyro_*.
 * Returns: nothing
 */
void convertGyroData() {
	char FS_SEL = mpu_read(GYRO_CONFIG);
	double factor = 1 << FS_SEL;
	factor = 131/factor;

	gyro_x = ((double) (gyro_x_raw - gyro_x_offset)) / factor;
	gyro_y = ((double) (gyro_y_raw - gyro_y_offset)) / factor;
	gyro_z = ((double) (gyro_z_raw - gyro_z_offset)) / factor;
}

/*
 * Reads in the raw x, y, and z values of the accelerometer and puts them
 * in global variables accel_*_raw.
 *
 * NOTE: This function does NOT check the Data Ready bit before reading
 * accel data. It is up to the caller to do that before calling this.
 *
 * Returns: nothing
 */
void readRawAccelData() {
	accel_x_upper = mpu_read(ACCEL_XOUT_H);
    accel_x_lower = mpu_read(ACCEL_XOUT_L);
    accel_x_raw   = combine(accel_x_upper, accel_x_lower);
    
    accel_y_upper = mpu_read(ACCEL_YOUT_H);
    accel_y_lower = mpu_read(ACCEL_YOUT_L);
    accel_y_raw   = combine(accel_y_upper, accel_y_lower);
    
    accel_z_upper = mpu_read(ACCEL_ZOUT_H);
    accel_z_lower = mpu_read(ACCEL_ZOUT_L);
    accel_z_raw   = combine(accel_z_upper, accel_z_lower);
}

//Converts the already acquired accelerometer data into 3D euler angles
void convertAccel2Angle() {
	x_angle = 57.295*atan((float) accel_y_raw / sqrt(pow((float)accel_z_raw,2)+pow((float)accel_x_raw,2)));
	y_angle = 57.295*atan((float)-accel_x_raw / sqrt(pow((float)accel_z_raw,2)+pow((float)accel_y_raw,2)));	
}

/*
 * Wakes up the MPU-5060 and initializes it with the proper settings.
 * Returns: nothing
 */
void mpu_init(void) {
	unsigned char whoami;
	unsigned char fail;
	unsigned char sleep;

	if (DEBUG) fprintf(stdout, "Initializing MPU...\n");

	//Reset all registers on the MPU
	fail = mpu_write(PWR_MGMT_1, 0b10000000);
	if (fail && DEBUG) fprintf(stderr, "MPU reset failed. failbit: %d\n", fail);
	
	_delay_ms(100);

	//Read and print the "Who am I" register
    if (DEBUG) {
	    whoami = mpu_read(WHO_AM_I);
	    fprintf(stdout, "who am I: 0x%x\n", whoami);
    }

	//Enable the Data Ready interrupt
	fail = mpu_write(INT_ENABLE, 0b1); //bit 0 is DATA_RDY_EN
    if (fail && DEBUG) fprintf(stderr, "DATA_RDY_EN not set. failbit: %d\n", fail);
	
	//Wake the MPU from sleep mode
	fail = mpu_write(PWR_MGMT_1, 0);
	if (fail && DEBUG) fprintf(stderr, "wake-up failed. failbit: %d\n", fail);

	//Check that the MPU has been awoken
	sleep = mpu_read(PWR_MGMT_1);
	if ((sleep == 0) && DEBUG) fprintf(stdout, "MPU is awake!\n");
	
	//Give the MPU some time to adjust
	_delay_ms(100);
	
	//Wait until the MPU has data ready
	while ((data_ready & 0b1) != 1) {       //bit 0 is DAT_RDY_INT
		data_ready = mpu_read(INT_STATUS);  //Check if data is ready
	}
	
	//Calibrate gyro
	readRawGyroData();
	//gyro offsets are now in gyro_*raw
	
	//put the read values into offset
	gyro_x_offset = gyro_x_raw;
	gyro_y_offset = gyro_y_raw;
	gyro_z_offset = gyro_z_raw;
}

/*
 * Initializes UART, I2C, and the necessary pins of the MCU.
 * Returns: nothing
 */
void initialize(void) {
	DDRB  = 0xff;    // use all pins on port B for output
	PORTB = 0x00;    // (active low LEDs)

	//set up INT0
    EIMSK = 1 << INT0 ; // turn on int0
    EICRA = 0b01;       // generate interrupt on either edge, pg 327, 65

	//init the UART
	uart_init();
	stdout = stdin = stderr = &uart_str;
	if (DEBUG) fprintf(stdout, "\n\nUART initialized!\n\r");

	//init I2C interface
	i2c_init();
	_delay_ms(100);

	sei();  // Interrupts - leggo!
}

//*********** External Interrupt ISR ***************
ISR (INT0_vect) {
	//check if D2 is HIGH or LOW
	if (check_bit(PIND, 2)) {
		//if D2 is high, then button is pressed
		buttonPressed = 1;
	} else {
		buttonPressed = 0;
	}
}

//*************** MAIN *****************
int main(void)
{
	initialize();
	mpu_init();
	
	//IMPORTANT: Make sure that twimaster.c is NOT in this project's folder.

    while(1)
    {
		if (buttonPressed == 0) {
			data_ready = mpu_read(INT_STATUS);  //Check if data is ready
			if ((data_ready & 0b1) == 1) {      //bit 0 is DAT_RDY_INT
			
				//readRawGyroData();
				readRawAccelData();

	            //Convert raw data to degrees
				//convertGyroData();
				convertAccel2Angle();

	            if (DEBUG) fprintf(stdout, "angles (x,y): (%d, %d)\n", (int)x_angle, (int)y_angle);
	            //This will get sent over Bluetooth to the other MCU
	            else fprintf(stdout, "%d,%d\n", (int)x_angle, (int)y_angle);
			}
		} else {
			fprintf(stdout, "%d,%d\n", 0, 0);
		}
        _delay_ms(200);
    }
}
